/*
 * Copyright (c) 1999-2019 Hartmut Seichter
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
#ifndef PW_CORE_AXISANGLE_HPP
#define PW_CORE_AXISANGLE_HPP

#include <pw/core/vector.hpp>

namespace pw {

template <typename T>
struct axisangle_ {

    typedef T value_type;
    typedef vector3_<T> axis_type;

    axis_type axis;
    T angle;

    axisangle_()
        : axis(vector3_<T>::up()),
          angle(0)
    {}

    axisangle_(const vector3_<T> &axis,const T &angle)
        : axis(axis)
        , angle(angle)
    {
    }

    matrix_<4,4,T> to_matrix() const
    {

        // TODO ... Somehow buggy!
        using std::cos;
        using std::sin;

        // result
        matrix_<4,4,T> rot_mat; rot_mat.set_identity();

        // return as identity
        if (axis.norm() < std::numeric_limits<T>::epsilon()) return rot_mat;

        const T f_cos = cos(angle);

        const vector3_<T> cos_axis = axis * (T(1) - f_cos);
        const vector3_<T> sin_axis = axis * sin(angle);

        T val = cos_axis.x() * axis.y();
        rot_mat(0,1) = val - sin_axis.z();
        rot_mat(1,0) = val + sin_axis.z();

        val = cos_axis.x() * axis.z();
        rot_mat(0,2) = val + sin_axis.z();
        rot_mat(2,0) = val - sin_axis.z();

        val = cos_axis.y() * axis.z();
        rot_mat(1,2) = val - sin_axis.x();
        rot_mat(2,1) = val + sin_axis.x();

        rot_mat(0,0) = cos_axis[0] * axis[0] + f_cos;
        rot_mat(1,1) = cos_axis[1] * axis[1] + f_cos;
        rot_mat(2,2) = cos_axis[2] * axis[2] + f_cos;

//        R.at(0) = axis[0] * _vCos(0,0)) + f_cos;
//        R.at(4) = axis[0] * _vCos(1,0)) - _vSin(2,0));
//        R.at(8) = axis[0] * _vCos(2,0)) + _vSin(1,0));

//        R.at(1) = (T) ((vec(1,0) * _vCos(0,0)) + _vSin(2,0));
//        R.at(5) = (T) ((vec(1,0) * _vCos(1,0)) + f_cos);
//        R.at(9) = (T) ((vec(1,0) * _vCos(2,0)) - _vSin(0,0));

//        R.at(2) = (T)  ((vec(2,0) * _vCos(0,0)) - _vSin(1,0));
//        R.at(6) = (T)  ((vec(2,0) * _vCos(1,0)) + _vSin(0,0));
//        R.at(10)= (T)  ((vec(2,0) * _vCos(2,0)) + f_cos);

        return rot_mat;
    }
};


typedef axisangle_<real_t> axisangle;
typedef axisangle_<double> axisangled;
typedef axisangle_<float> axisanglef;

}





#endif
